"""Module containing custom animation classes."""

from .cursormoveanimation import CursorMoveAnimation
from .slideanimation import SlideAnimation
from .textanimation import TextAnimation
from .zoomanimation import ZoomAnimation


__all__ = ["SlideAnimation", "ZoomAnimation", "TextAnimation", "CursorMoveAnimation"]
